import ast
from typing import Iterable
from typing import Tuple

from tokenize_rt import Offset

from pyupgrade._ast_helpers import ast_to_offset
from pyupgrade._data import register
from pyupgrade._data import State
from pyupgrade._data import TokenFunc
from pyupgrade._token_helpers import replace_list_comp_brackets


@register(ast.Starred)
def visit_Starred(
        state: State,
        node: ast.Starred,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if (
        state.settings.min_version >= (3,) and
        isinstance(node.value, ast.ListComp)
    ):
        yield ast_to_offset(node.value), replace_list_comp_brackets



import ast
from typing import Iterable
from typing import List
from typing import Tuple

from tokenize_rt import Offset
from tokenize_rt import Token

from pyupgrade._ast_helpers import ast_to_offset
from pyupgrade._data import register
from pyupgrade._data import State
from pyupgrade._data import TokenFunc
from pyupgrade._token_helpers import find_token


def _replace_celementtree_with_elementtree(
        i: int,
        tokens: List[Token],
) -> None:
    j = find_token(tokens, i, 'cElementTree')
    tokens[j] = tokens[j]._replace(src='ElementTree')


@register(ast.ImportFrom)
def visit_ImportFrom(
        state: State,
        node: ast.ImportFrom,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if (
            state.settings.min_version >= (3,) and
            node.module == 'xml.etree.cElementTree' and
            node.level == 0
    ):
        yield ast_to_offset(node), _replace_celementtree_with_elementtree


@register(ast.Import)
def visit_Import(
        state: State,
        node: ast.Import,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if (
        state.settings.min_version >= (3,) and
        len(node.names) == 1 and
        node.names[0].name == 'xml.etree.cElementTree' and
        node.names[0].asname is not None
    ):
        yield ast_to_offset(node), _replace_celementtree_with_elementtree



import ast
from typing import Iterable
from typing import List
from typing import Tuple

from tokenize_rt import NON_CODING_TOKENS
from tokenize_rt import Offset
from tokenize_rt import Token

from pyupgrade._ast_helpers import ast_to_offset
from pyupgrade._ast_helpers import contains_await
from pyupgrade._data import register
from pyupgrade._data import State
from pyupgrade._data import TokenFunc
from pyupgrade._token_helpers import find_closing_bracket
from pyupgrade._token_helpers import find_comprehension_opening_bracket
from pyupgrade._token_helpers import replace_list_comp_brackets


ALLOWED_FUNCS = frozenset((
    'bytearray',
    'bytes',
    'frozenset',
    'list',
    'max',
    'min',
    'sorted',
    'sum',
    'tuple',
))


def _delete_list_comp_brackets(i: int, tokens: List[Token]) -> None:
    start = find_comprehension_opening_bracket(i, tokens)
    end = find_closing_bracket(tokens, start)
    tokens[end] = Token('PLACEHOLDER', '')
    tokens[start] = Token('PLACEHOLDER', '')
    j = end + 1
    while j < len(tokens) and tokens[j].name in NON_CODING_TOKENS:
        j += 1
    if tokens[j].name == 'OP' and tokens[j].src == ',':
        tokens[j] = Token('PLACEHOLDER', '')


@register(ast.Call)
def visit_Call(
        state: State,
        node: ast.Call,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if (
            isinstance(node.func, ast.Name) and
            node.func.id in ALLOWED_FUNCS and
            node.args and
            isinstance(node.args[0], ast.ListComp) and
            not any(
                generator.is_async
                for generator in node.args[0].generators
            ) and
            not contains_await(node.args[0])
    ):
        if len(node.args) == 1 and not node.keywords:
            yield ast_to_offset(node.args[0]), _delete_list_comp_brackets
        else:
            yield ast_to_offset(node.args[0]), replace_list_comp_brackets



import ast
from typing import Iterable
from typing import Tuple

from tokenize_rt import List
from tokenize_rt import NON_CODING_TOKENS
from tokenize_rt import Offset
from tokenize_rt import Token

from pyupgrade._ast_helpers import ast_to_offset
from pyupgrade._data import register
from pyupgrade._data import State
from pyupgrade._data import TokenFunc
from pyupgrade._token_helpers import find_closing_bracket


def _replace_dict_brackets(i: int, tokens: List[Token]) -> None:
    closing = find_closing_bracket(tokens, i)
    j = closing - 1
    while tokens[j].name in NON_CODING_TOKENS and j > i:
        j -= 1
    if tokens[j].name == 'OP' and tokens[j].src == ',':
        tokens[j] = Token('PLACEHOLDER', '')

    if tokens[i].line == tokens[closing].line:
        tokens[i] = Token('PLACEHOLDER', '')
        tokens[closing] = Token('PLACEHOLDER', '')
    else:
        tokens[i] = Token('CODE', '(')
        tokens[closing] = Token('CODE', ')')


def _remove_double_star(i: int, tokens: List[Token]) -> None:
    j = i
    while not (tokens[j].name == 'OP' and tokens[j].src == '**'):
        j -= 1
    tokens[j] = Token('PLACEHOLDER', '')


def _replace_comma_with_pipe(i: int, tokens: List[Token]) -> None:
    j = i - 1
    while not (tokens[j].name == 'OP' and tokens[j].src == ','):
        j -= 1
    tokens[j] = Token('CODE', ' |')


@register(ast.Dict)
def visit_Dict(
        state: State,
        node: ast.Dict,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if state.settings.min_version < (3, 9):
        return

    if all(key is None for key in node.keys) and len(node.values) > 1:
        yield ast_to_offset(node), _replace_dict_brackets
        for idx, arg in enumerate(node.values):
            yield ast_to_offset(arg), _remove_double_star
            if idx > 0:
                yield ast_to_offset(arg), _replace_comma_with_pipe



import ast
import functools
from typing import Iterable
from typing import List
from typing import Tuple

from tokenize_rt import Offset
from tokenize_rt import Token

from pyupgrade._ast_helpers import ast_to_offset
from pyupgrade._data import register
from pyupgrade._data import State
from pyupgrade._data import TokenFunc
from pyupgrade._token_helpers import find_closing_bracket
from pyupgrade._token_helpers import find_open_paren

NUM_TYPES = {
    int: 'int',
    float: 'float',
    complex: 'complex',
}


def _rewrite_type_of_primitive(
        i: int,
        tokens: List[Token],
        *,
        src: str,
) -> None:
    open_paren = find_open_paren(tokens, i + 1)
    j = find_closing_bracket(tokens, open_paren)
    tokens[i] = tokens[i]._replace(src=src)
    del tokens[i + 1:j + 1]


@register(ast.Call)
def visit_Call(
        state: State,
        node: ast.Call,
        parent: ast.AST,
) -> Iterable[Tuple[Offset, TokenFunc]]:
    if (
            state.settings.min_version >= (3,) and
            isinstance(node.func, ast.Name) and
            node.func.id == 'type' and
            len(node.args) == 1
    ):
        if isinstance(node.args[0], ast.Str):
            func = functools.partial(
                _rewrite_type_of_primitive,
                src='str',
            )
            yield ast_to_offset(node), func
        elif isinstance(node.args[0], ast.Bytes):
            func = functools.partial(
                _rewrite_type_of_primitive,
                src='bytes',
            )
            yield ast_to_offset(node), func
        elif isinstance(node.args[0], ast.Num):
            func = functools.partial(
                _rewrite_type_of_primitive,
                src=NUM_TYPES[type(node.args[0].n)],
            )
            yield ast_to_offset(node), func
